#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gcl_iam import middlewares as iam_mw
from restalchemy.api import applications
from restalchemy.api.middlewares import logging as logging_mw
from restalchemy.api import middlewares
from restalchemy.api import routes
from restalchemy.openapi import structures as openapi_structures
from restalchemy.openapi import engines as openapi_engines

from genesis_notification.common.api.middlewares import errors as errors_mw
from genesis_notification.user_api.api import routes as app_routes
from genesis_notification.user_api.api import versions
from genesis_notification import version as app_version


skip_auth_endpoints = [
    iam_mw.EndpointComparator("/"),
    iam_mw.EndpointComparator("/v1/"),
]


class UserApiApp(routes.RootRoute):
    pass


# Route to /v1/ endpoint.
setattr(
    UserApiApp,
    versions.API_VERSION_1_0,
    routes.route(app_routes.ApiEndpointRoute),
)


def get_api_application():
    return UserApiApp


def get_openapi_engine():
    openapi_engine = openapi_engines.OpenApiEngine(
        info=openapi_structures.OpenApiInfo(
            title=f"Genesis Notification {versions.API_VERSION_1_0} User API",
            version=app_version.version_info,
            description=(f"OpenAPI - Genesis Notification {versions.API_VERSION_1_0}"),
        ),
        paths=openapi_structures.OpenApiPaths(),
        components=openapi_structures.OpenApiComponents(),
    )
    return openapi_engine


def build_wsgi_application(iam_engine_driver):
    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=get_api_application(),
            openapi_engine=get_openapi_engine(),
        ),
        [
            middlewares.configure_middleware(
                iam_mw.GenesisCoreAuthMiddleware,
                iam_engine_driver=iam_engine_driver,
                skip_auth_endpoints=skip_auth_endpoints,
            ),
            errors_mw.ErrorsHandlerMiddleware,
            logging_mw.LoggingMiddleware,
        ],
    )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import logging
from email.mime import text
from email.mime import multipart
import smtplib

import jinja2
from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.dm import types_dynamic
from restalchemy.storage.sql import orm
import zulip

from genesis_notification.common import constants as c


LOG = logging.getLogger(__name__)


def next_time(seconds):

    def calulator():
        now = datetime.datetime.now(datetime.timezone.utc)
        delta = datetime.timedelta(seconds=seconds)
        return now + delta

    return calulator


class ModelWithAlwaysActiveStatus(models.Model):
    STATUS = c.AlwaysActiveStatus

    status = properties.property(
        types.Enum([s.value for s in c.AlwaysActiveStatus]),
        default=STATUS.ACTIVE.value,
    )


class SimpleSmtpProtocol(types_dynamic.AbstractKindModel):
    KIND = "SimpleSMTP"

    host = properties.property(
        types.String(min_length=1, max_length=128),
        required=True,
    )
    port = properties.property(
        types.Integer(min_value=1, max_value=65535),
        required=True,
    )
    noreply_email_address = properties.property(
        types.Email(),
        required=True,
    )

    def _build_message(self, content, user_context):
        msg = multipart.MIMEMultipart("alternative")
        msg["From"] = self.noreply_email_address
        msg["To"] = user_context["user"]["email"]
        msg["Subject"] = content.title
        for body in content.bodies:
            msg.attach(text.MIMEText(body, "html", "utf-8"))
        return msg

    def _authenticate(self, smtp):
        return smtp

    def send(self, content, user_context):
        msg = self._build_message(content, user_context)
        with smtplib.SMTP(self.host, self.port) as smtp:
            smtp = self._authenticate(smtp)
            return smtp.sendmail(
                from_addr=self.noreply_email_address,
                to_addrs=user_context["user"]["email"],
                msg=msg.as_string(),
            )


class StartTlsSmtpProtocol(SimpleSmtpProtocol):
    KIND = "StartTlsSMTP"

    user = properties.property(
        types.Email(),
        required=True,
    )
    password = properties.property(
        types.String(max_length=128, min_length=1),
        required=True,
    )

    def _authenticate(self, smtp):
        smtp.ehlo()
        smtp.starttls()
        smtp.ehlo()
        smtp.login(self.user, self.password)
        return smtp


class ZulipProtocol(types_dynamic.AbstractKindModel):
    KIND = "zulip"

    endpoint = properties.property(
        types.Url(),
        required=True,
    )
    email_address = properties.property(
        types.Email(),
        required=True,
    )
    api_key = properties.property(
        types.String(),
        required=True,
    )

    def send(self, content, user_context):
        client = zulip.Client(
            site=self.endpoint,
            email=self.email_address,
            api_key=self.api_key,
        )
        if content.KIND == RenderedStreamMessageContent.KIND:
            request = {
                "type": "stream",
                "to": content.to,
                "topic": content.topic,
                "content": content.content,
            }
        elif content.KIND == RenderedDirectMessageContent.KIND:
            request = {
                "type": "direct",
                "to": content.to,
                "content": content.content,
            }
        else:
            raise NotImplementedError(f"Unsupported content type {content}")
        result = client.send_message(request)
        if result["result"] != "success":
            raise RuntimeError(f"Failed to send message: {result['msg']}")


class Provider(
    models.ModelWithUUID,
    models.ModelWithRequiredNameDesc,
    ModelWithAlwaysActiveStatus,
    models.ModelWithProject,
    models.ModelWithTimestamp,
    orm.SQLStorableMixin,
):
    __tablename__ = "providers"

    protocol = properties.property(
        types_dynamic.KindModelSelectorType(
            types_dynamic.KindModelType(SimpleSmtpProtocol),
            types_dynamic.KindModelType(StartTlsSmtpProtocol),
            types_dynamic.KindModelType(ZulipProtocol),
        ),
        required=True,
    )

    def send(self, content, user_context):
        return self.protocol.send(
            content=content,
            user_context=user_context,
        )


class EventType(
    models.ModelWithUUID,
    models.ModelWithNameDesc,
    ModelWithAlwaysActiveStatus,
    models.ModelWithProject,
    models.ModelWithTimestamp,
    orm.SQLStorableMixin,
):
    __tablename__ = "event_types"


class AbstractContent(types_dynamic.AbstractKindModel):
    def get_id(self):
        return "%s" % self.__class__.__name__.lower()


class RenderedEmailContent(AbstractContent):
    KIND = "rendered_email"

    title = properties.property(
        types.String(max_length=256),
        default="",
    )
    bodies = properties.property(
        types.List(),
        default=list,
    )


class EmailContent(RenderedEmailContent):
    KIND = "email"

    def render(self, params):
        return RenderedEmailContent(
            title=jinja2.Template(self.title).render(**params),
            bodies=[jinja2.Template(body).render(**params) for body in self.bodies],
        )


class RenderedStreamMessageContent(AbstractContent):
    KIND = "rendered_zulip_stream_message"

    to = properties.property(
        types.String(min_length=1, max_length=256),
        required=True,
    )
    topic = properties.property(
        types.String(min_length=1, max_length=256),
        required=True,
    )
    content = properties.property(
        types.String(min_length=1, max_length=10000),
        required=True,
    )


class ZulipStreamMessageContent(RenderedStreamMessageContent):
    KIND = "zulip_stream_message"

    to = properties.property(
        types.String(min_length=1, max_length=256),
        default="{{ channel }}",
    )
    topic = properties.property(
        types.String(min_length=1, max_length=256),
        default="{{ topic }}",
    )
    content = properties.property(
        types.String(min_length=1, max_length=10000),
        default="{{ message }}",
    )

    def render(self, params):
        return RenderedStreamMessageContent(
            to=jinja2.Template(self.to).render(**params),
            topic=jinja2.Template(self.topic).render(**params),
            content=jinja2.Template(self.content).render(**params),
        )


class RenderedDirectMessageContent(AbstractContent):
    KIND = "rendered_zulip_direct_message"

    to = properties.property(
        types.TypedList(nested_type=types.String(min_length=1, max_length=256)),
        required=True,
    )
    content = properties.property(
        types.String(min_length=1, max_length=10000),
        required=True,
    )


class ZulipDirectMessageContent(RenderedDirectMessageContent):
    KIND = "zulip_direct_message"

    to = properties.property(
        types.String(min_length=1, max_length=256),
        default="{{ users }}",
    )
    content = properties.property(
        types.String(min_length=1, max_length=10000),
        default="{{ message }}",
    )

    def render(self, params):
        return RenderedDirectMessageContent(
            to=[
                user.strip()
                for user in jinja2.Template(self.to).render(**params).split(",")
            ],
            content=jinja2.Template(self.content).render(**params),
        )


ZULIP_MESSAGE_TYPE_MAP = {
    RenderedStreamMessageContent.KIND: "stream",
    RenderedDirectMessageContent.KIND: "direct",
}


class Template(
    models.ModelWithUUID,
    models.ModelWithRequiredNameDesc,
    ModelWithAlwaysActiveStatus,
    models.ModelWithProject,
    models.ModelWithTimestamp,
    orm.SQLStorableMixin,
):
    __tablename__ = "templates"

    content = properties.property(
        types_dynamic.KindModelSelectorType(
            types_dynamic.KindModelType(EmailContent),
            types_dynamic.KindModelType(ZulipStreamMessageContent),
            types_dynamic.KindModelType(ZulipDirectMessageContent),
        ),
        required=True,
    )
    params = properties.property(
        types.Dict(),
        required=True,
    )
    provider = relationships.relationship(
        Provider,
        required=True,
        prefetch=True,
    )
    event_type = relationships.relationship(
        EventType,
        required=True,
    )
    is_default = properties.property(
        types.Boolean(),
        default=False,
    )


class UserExchange(types_dynamic.AbstractKindModel):
    KIND = "User"

    user_id = properties.property(
        types.UUID(),
        required=True,
    )

    def get_context(self, iam_client):
        return {
            "user": iam_client.get_user(self.user_id),
        }


class DummyExchange(types_dynamic.AbstractKindModel):
    KIND = "Dummy"

    def get_context(self, iam_client):
        return {}


class ProjectExchange(types_dynamic.AbstractKindModel):
    KIND = "Project"

    project_id = properties.property(
        types.UUID(),
        required=True,
    )


class SystemExchange(types_dynamic.AbstractKindModel):
    KIND = "System"


class Binding(
    models.ModelWithUUID,
    models.ModelWithProject,
    ModelWithAlwaysActiveStatus,
    models.ModelWithTimestamp,
    orm.SQLStorableMixin,
):
    user = properties.property(
        types.UUID(),
        required=True,
    )
    template = relationships.relationship(
        Template,
        required=True,
    )
    event_type = relationships.relationship(
        EventType,
        required=True,
    )


class StatusMixin(models.Model):
    next_retry_delta = 60  # 60 sec
    last_retry_delta = 1 * 24 * 60 * 60  # 1 day

    STATUS = c.EventStatus

    status = properties.property(
        types.Enum([s.value for s in STATUS]),
        default=STATUS.NEW.value,
    )
    status_description = properties.property(
        types.String(max_length=255),
        default="",
    )
    next_retry_at = properties.property(
        types.UTCDateTimeZ(),
        default=next_time(seconds=0),  # now + 0sec
    )
    last_retry_at = properties.property(
        types.UTCDateTimeZ(),
        default=next_time(seconds=last_retry_delta),  # now + 1day
    )
    retry_count = properties.property(
        types.Integer(min_value=0, max_value=65536),
        default=0,
    )

    def reset_next_retry(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        self.next_retry_at = now + datetime.timedelta(seconds=self.next_retry_delta)

    def set_error_status(self, error_message):
        self.status_description = str(error_message)
        self.retry_count += 1
        self.status = self.STATUS.ERROR.value
        self.reset_next_retry()
        self.save()

    def set_done_status(self):
        self.status_description = "ok"
        self.status = self.STATUS.ACTIVE.value


class Event(
    models.ModelWithUUID,
    models.ModelWithNameDesc,
    models.ModelWithProject,
    models.ModelWithTimestamp,
    StatusMixin,
    orm.SQLStorableMixin,
):
    __tablename__ = "events"

    exchange = properties.property(
        types_dynamic.KindModelSelectorType(
            types_dynamic.KindModelType(UserExchange),
            types_dynamic.KindModelType(ProjectExchange),
            types_dynamic.KindModelType(SystemExchange),
            types_dynamic.KindModelType(DummyExchange),
        ),
        default=DummyExchange,
    )
    event_params = properties.property(
        types.Dict(),
        required=True,
    )
    event_type = relationships.relationship(
        EventType,
        required=True,
    )

    def get_context(self, iam_client):
        return self.exchange.get_context(iam_client)

    def render(self, iam_client):
        context = self.get_context(iam_client)
        templates = Template.objects.get_all(
            filters={
                "event_type": filters.EQ(self.event_type),
                "is_default": filters.EQ(True),
            }
        )

        rendered_events = []
        for template in templates:
            params = context.copy()
            params.update(self.event_params)
            rendered_content = template.content.render(params)
            rendered_events.append(
                RenderedEvent(
                    content=rendered_content,
                    event_id=self.uuid,
                    provider=template.provider,
                    user_context=context,
                    status=StatusMixin.STATUS.IN_PROGRESS.value,
                )
            )

        return rendered_events


class RenderedEvent(
    models.ModelWithUUID,
    models.ModelWithTimestamp,
    StatusMixin,
    orm.SQLStorableMixin,
):
    __tablename__ = "rendered_events"

    content = properties.property(
        types_dynamic.KindModelSelectorType(
            types_dynamic.KindModelType(RenderedEmailContent),
            types_dynamic.KindModelType(RenderedStreamMessageContent),
            types_dynamic.KindModelType(RenderedDirectMessageContent),
        ),
        required=True,
    )
    event_id = properties.property(
        types.UUID(),
        required=True,
    )
    provider = relationships.relationship(
        Provider,
        required=True,
        prefetch=True,
    )
    user_context = properties.property(
        types.Dict(),
        required=True,
    )

    def send(self):
        try:
            self.provider.send(
                content=self.content,
                user_context=self.user_context,
            )
        except Exception as e:
            LOG.exception("Failed to send event")
            self.set_error_status(e)
            return
        self.set_done_status()
        self.update()


class UnprocessedEvent(
    models.ModelWithUUID,
    orm.SQLStorableMixin,
):
    __tablename__ = "unprocessed_events"
    event = relationships.relationship(
        Event,
        required=True,
        prefetch=True,
    )
    next_retry_at = properties.property(
        types.UTCDateTimeZ(),
        required=True,
    )
    last_retry_at = properties.property(
        types.UTCDateTimeZ(),
        required=True,
    )

    def process_event(self, iam_client):
        try:
            rendered_events = self.event.render(iam_client)
        except Exception as e:
            LOG.exception("Failed to render event")
            self.event.set_error_status(e)
            return
        for rendered_event in rendered_events:
            rendered_event.insert()


class IncorrectStatuses(
    models.ModelWithUUID,
    orm.SQLStorableMixin,
):
    __tablename__ = "incorrect_statuses"

    STATUS = c.EventStatus

    event = relationships.relationship(
        Event,
        required=True,
        prefetch=True,
    )
    user_status = properties.property(
        types.Enum([s.value for s in STATUS]),
        required=True,
    )
    user_status_description = properties.property(
        types.String(),
        default="",
    )
    system_status = properties.property(
        types.Enum([s.value for s in STATUS]),
        required=True,
    )
    system_status_description = properties.property(
        types.String(),
        default="",
    )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import logging

from gcl_looper.services import basic
from restalchemy.common import contexts
from restalchemy.dm import filters

from genesis_notification.dm import models


LOG = logging.getLogger(__name__)


class EventBuilderAgent(basic.BasicService):
    def __init__(self, iam_client, butch_size=100, **kwargs):
        self._iam_client = iam_client
        self._butch_size = butch_size
        super().__init__(**kwargs)

    def _setup(self):
        pass

    def _process_unprocessed_events(self):
        unprocessed_events = models.UnprocessedEvent.objects.get_all(
            filters={
                "next_retry_at": filters.LT(
                    datetime.datetime.now(tz=datetime.timezone.utc)
                ),
            },
            limit=self._butch_size,
        )
        for e in unprocessed_events:
            e.process_event(iam_client=self._iam_client)

    def _sync_event_statuses(self):
        for item in models.IncorrectStatuses.objects.get_all(
            limit=self._butch_size,
        ):
            LOG.info("Syncing item status: %r", item)
            event = item.event
            event.status = item.system_status
            event.status_description = item.system_status_description
            event.update()

    def _cleanup(self):
        events = models.Event.objects.get_all(
            filters={
                "last_retry_at": filters.LT(
                    datetime.datetime.now(tz=datetime.timezone.utc)
                )
            },
            limit=self._butch_size,
        )
        rendered_events = models.RenderedEvent.objects.get_all(
            filters={
                "last_retry_at": filters.LT(
                    datetime.datetime.now(tz=datetime.timezone.utc)
                )
            },
            limit=self._butch_size,
        )

        for event in events + rendered_events:
            LOG.info("Cleaning up event: %r", event)
            event.delete()

    def _iteration(self):
        ctx = contexts.Context()
        with ctx.session_manager():
            self._process_unprocessed_events()
            self._sync_event_statuses()
            self._cleanup()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import logging

from gcl_looper.services import basic
from restalchemy.common import contexts
from restalchemy.dm import filters

from genesis_notification.dm import models


LOG = logging.getLogger(__name__)


class SMTPAgent(basic.BasicService):
    def __init__(self, butch_size=100, iter_min_period=5, **kwargs):
        self._butch_size = butch_size
        super().__init__(iter_min_period=iter_min_period, **kwargs)

    def _setup(self):
        pass

    def _process_events(self):
        for event in models.RenderedEvent.objects.get_all(
            filters={
                "status": filters.NE(models.RenderedEvent.STATUS.ACTIVE.value),
                "next_retry_at": filters.LT(
                    datetime.datetime.now(tz=datetime.timezone.utc)
                ),
            },
            limit=self._butch_size,
        ):
            LOG.info("Processing event: %s", event)
            event.send()

    def _iteration(self):
        ctx = contexts.Context()
        with ctx.session_manager():
            self._process_events()



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
# Copyright 2025 Genesis Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstarctMigrationStep):
    def __init__(self):
        self._depends = []

    @property
    def migration_id(self):
        return "40a307b3-fdcc-46d8-bc81-1e2a53ac59e4"

    @property
    def is_manual(self):
        return False

    def upgrade(self, session):
        expressions = [
            """
            CREATE TYPE "enum_status_active" AS ENUM ('ACTIVE');
            """,
            """
            CREATE TYPE "enum_event_status" AS ENUM (
                'NEW',
                'IN_PROGRESS',
                'ACTIVE',
                'ERROR'
            );
            """,
            """
            CREATE TABLE "providers" (
                "uuid" CHAR(36) PRIMARY KEY,
                "name" VARCHAR(256) NOT NULL,
                "description" VARCHAR(256) NOT NULL,
                "status" enum_status_active NOT NULL DEFAULT 'ACTIVE',
                "project_id" CHAR(36) NOT NULL,
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "protocol" JSONB NOT NULL
            );
            """,
            """
            CREATE INDEX "providers_project_idx" ON "providers" (project_id);
            """,
            """
            CREATE INDEX "providers_name_idx" ON "providers" (name);
            """,
            """
            CREATE TABLE "event_types" (
                "uuid" CHAR(36) PRIMARY KEY,
                "name" VARCHAR(256) NOT NULL DEFAULT '',
                "description" VARCHAR(256) NOT NULL DEFAULT '',
                "status" enum_status_active NOT NULL DEFAULT 'ACTIVE',
                "project_id" CHAR(36),
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW()
            );
            """,
            """
            CREATE INDEX "event_types_project_idx" ON "event_types" (
                "project_id"
            );
            """,
            """
            CREATE TABLE "templates" (
                "uuid" CHAR(36) PRIMARY KEY,
                "name" VARCHAR(256) NOT NULL,
                "description" VARCHAR(256) NOT NULL,
                "status" enum_status_active NOT NULL DEFAULT 'ACTIVE',
                "project_id" CHAR(36) NOT NULL,
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "content" JSONB NOT NULL,
                "params" JSONB NOT NULL,
                "provider" CHAR(36) NOT NULL REFERENCES providers(uuid),
                "event_type" CHAR(36) NOT NULL REFERENCES event_types(uuid),
                "is_default" BOOLEAN NOT NULL DEFAULT FALSE
            );
            """,
            """
            CREATE INDEX "templates_project_idx" ON "templates" (
                "project_id"
            );
            """,
            """
            CREATE INDEX "templates_provider_idx" ON "templates" ("provider");
            """,
            """
            CREATE INDEX "templates_event_type_idx" ON "templates" (
                "event_type"
            );
            """,
            """
            CREATE TABLE "bindings" (
                "uuid" CHAR(36) PRIMARY KEY,
                "status" enum_status_active NOT NULL DEFAULT 'ACTIVE',
                "project_id" CHAR(36) NOT NULL,
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "user" CHAR(36) NOT NULL,
                "template" CHAR(36) NOT NULL REFERENCES templates(uuid),
                "event_type" CHAR(36) NOT NULL REFERENCES event_types(uuid)
            );
            """,
            """
            CREATE INDEX "bindings_user_idx" ON "bindings" ("user");
            """,
            """
            CREATE INDEX "bindings_template_idx" ON "bindings" ("template");
            """,
            """
            CREATE INDEX "bindings_event_type_idx" ON "bindings" (
                "event_type"
            );
            """,
            """
            CREATE TABLE "events" (
                "uuid" CHAR(36) PRIMARY KEY,
                "name" VARCHAR(256) NOT NULL DEFAULT '',
                "description" VARCHAR(256) NOT NULL DEFAULT '',
                "project_id" CHAR(36) NOT NULL,
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "status" enum_event_status NOT NULL DEFAULT 'NEW',
                "status_description" VARCHAR(256) NOT NULL DEFAULT '',
                "next_retry_at" TIMESTAMP(6) NOT NULL,
                "last_retry_at" TIMESTAMP(6) NOT NULL,
                "retry_count" INTEGER NOT NULL DEFAULT 0,
                "exchange" JSONB NOT NULL,
                "event_params" JSONB NOT NULL,
                "event_type" CHAR(36) NOT NULL REFERENCES event_types(uuid)
            );
            """,
            """
            CREATE INDEX "events_status_idx" ON "events" (status);
            """,
            """
            CREATE INDEX "events_event_type_idx" ON "events" (event_type);
            """,
            """
            CREATE TABLE "rendered_events" (
                "uuid" CHAR(36) PRIMARY KEY,
                "created_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "updated_at" TIMESTAMP(6) NOT NULL DEFAULT NOW(),
                "status" enum_event_status NOT NULL DEFAULT 'IN_PROGRESS',
                "status_description" VARCHAR(256) NOT NULL DEFAULT '',
                "next_retry_at" TIMESTAMP(6) NOT NULL,
                "last_retry_at" TIMESTAMP(6) NOT NULL,
                "retry_count" INTEGER NOT NULL DEFAULT 0,
                "content" JSONB NOT NULL,
                "event_id" CHAR(36) NOT NULL,
                "provider" CHAR(36) NOT NULL REFERENCES providers(uuid),
                "user_context" JSONB NOT NULL
            );
            """,
            """
            CREATE INDEX "rendered_events_status_idx" ON "rendered_events" (
                "status"
            );
            """,
            """
            CREATE INDEX "rendered_events_provider_idx" ON "rendered_events" (
                "provider"
            );
            """,
            """
            CREATE OR REPLACE VIEW unprocessed_events AS
                SELECT
                    e.uuid AS uuid,
                    e.uuid AS event,
                    e.next_retry_at AS next_retry_at,
                    e.last_retry_at AS last_retry_at
                FROM
                    events e
                LEFT JOIN
                    rendered_events re ON (
                        re.event_id = e.uuid
                    )
                WHERE
                    re.uuid IS NULL;
            """,
            """
            CREATE OR REPLACE VIEW "incorrect_statuses" as
                SELECT
                    e."uuid",
                    e."uuid" as "event",
                    e."status" as "user_status",
                    e."status_description" as "user_status_description",
                    re."status" as "system_status",
                    re."status_description" as "system_status_description"
                FROM
                    "events" e
                left JOIN
                    "rendered_events" re ON (
                        re."event_id" = e."uuid"
                    )
                WHERE
                    re."uuid" is NOT NULL AND
                    (
                        e."status" != re."status" OR
                        e."status_description" != re."status_description"
                    )
            """,
        ]

        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        views = ["unprocessed_events", "incorrect_statuses"]

        tables = [
            "rendered_events",
            "events",
            "bindings",
            "templates",
            "event_types",
            "providers",
        ]

        for view in views:
            self._delete_view_if_exists(session, view)

        for table in tables:
            self._delete_table_if_exists(session, table)

        session.execute('DROP TYPE IF EXISTS "enum_status_active" CASCADE;')
        session.execute('DROP TYPE IF EXISTS "enum_event_status" CASCADE;')


migration_step = MigrationStep()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import controllers as ra_controllers
from restalchemy.api import resources

from genesis_notification.dm import models
from genesis_notification.user_api.api import versions


class RootController(ra_controllers.Controller):
    """Controller for / endpoint"""

    def filter(self, filters):
        return (versions.API_VERSION_1_0,)


class ApiEndpointController(ra_controllers.RoutesListController):
    """Controller for /v1/ endpoint"""

    __TARGET_PATH__ = "/v1/"


class ProviderController(ra_controllers.BaseResourceController):
    __resource__ = resources.ResourceByRAModel(
        models.Provider, convert_underscore=False
    )


class TemplateController(ra_controllers.BaseResourceController):
    __resource__ = resources.ResourceByRAModel(
        models.Template, convert_underscore=False
    )


class EventTypeController(ra_controllers.BaseResourceController):
    __resource__ = resources.ResourceByRAModel(
        models.EventType, convert_underscore=False
    )


class EventController(ra_controllers.BaseResourceController):
    __resource__ = resources.ResourceByRAModel(models.Event, convert_underscore=False)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import sys

from gcl_looper.services import bjoern_service
from gcl_looper.services import hub
from gcl_iam import drivers
from gcl_iam import opts as iam_opts
from oslo_config import cfg
from restalchemy.common import config_opts as ra_config_opts
from restalchemy.storage.sql import engines

from genesis_notification.user_api.api import app
from genesis_notification.common import config
from genesis_notification.common import log as infra_log


api_cli_opts = [
    cfg.StrOpt(
        "bind-host",
        default="127.0.0.1",
        help="The host IP to bind to",
    ),
    cfg.IntOpt(
        "bind-port",
        default=8080,
        help="The port to bind to",
    ),
    cfg.IntOpt(
        "workers",
        default=1,
        help="How many http servers should be started",
    ),
]


DOMAIN = "user_api"

CONF = cfg.CONF
CONF.register_cli_opts(api_cli_opts, DOMAIN)
ra_config_opts.register_posgresql_db_opts(CONF)
iam_opts.register_iam_cli_opts(CONF)


def main():
    # Parse config
    config.parse(sys.argv[1:])

    # Configure logging
    infra_log.configure()
    log = logging.getLogger(__name__)

    log.info(
        "Start service on %s:%s",
        CONF[DOMAIN].bind_host,
        CONF[DOMAIN].bind_port,
    )

    service_hub = hub.ProcessHubService()
    iam_driver = drivers.HttpDriver(
        CONF.iam.iam_endpoint,
        CONF.iam.audience,
        CONF.iam.hs256_jwks_decryption_key,
    )

    for _ in range(CONF[DOMAIN].workers):
        service = bjoern_service.BjoernService(
            wsgi_app=app.build_wsgi_application(iam_driver),
            host=CONF[DOMAIN].bind_host,
            port=CONF[DOMAIN].bind_port,
            bjoern_kwargs=dict(reuse_port=True),
        )

        service.add_setup(
            lambda: engines.engine_factory.configure_postgresql_factory(conf=CONF)
        )

        service_hub.add_service(service)

    if CONF[DOMAIN].workers > 1:
        service_hub.start()
    else:
        service.start()
    log.info("Bye!!!")


if __name__ == "__main__":
    main()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from bazooka import common
from bazooka import client


class IAMClient(common.RESTClientMixIn):
    USER_PATH = "iam/users"

    def __init__(self, endpoint, token, timeout=5):
        super().__init__()
        self._client = client.Client(default_timeout=timeout)
        self._endpoint = endpoint
        self._token = token

    def _build_headers(self, **kwargs):
        headers = kwargs.copy()
        headers["Authorization"] = f"Bearer {self._token}"
        return headers

    def get_user(self, user_id):
        url = self._build_resource_uri([self.USER_PATH, user_id])
        return self._client.get(
            url,
            headers=self._build_headers(),
        ).json()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import sys

from oslo_config import cfg
from restalchemy.common import config_opts as ra_config_opts
from restalchemy.storage.sql import engines

from genesis_notification.clients import iam
from genesis_notification.common import config
from genesis_notification.common import log as infra_log
from genesis_notification.services.builders import agents

DOMAIN = "builder_agent"
DOMAIN_IAM_CLIENT = "iam_client"


iam_client_cli_opts = [
    cfg.StrOpt(
        "token",
        required=True,
        help="IAM token",
    ),
    cfg.URIOpt(
        "endpoint",
        default="http://127.0.0.1:11010/v1/",
        help="IAM endpoint",
    ),
    cfg.IntOpt(
        "timeout",
        default=5,
        min=0,
        help="IAM timeout",
    ),
]


CONF = cfg.CONF
ra_config_opts.register_posgresql_db_opts(CONF)
CONF.register_cli_opts(iam_client_cli_opts, DOMAIN_IAM_CLIENT)


def main():
    config.parse(sys.argv[1:])

    infra_log.configure()
    log = logging.getLogger(__name__)

    engines.engine_factory.configure_postgresql_factory(CONF)
    iam_client = iam.IAMClient(
        endpoint=CONF[DOMAIN_IAM_CLIENT].endpoint,
        token=CONF[DOMAIN_IAM_CLIENT].token,
        timeout=CONF[DOMAIN_IAM_CLIENT].timeout,
    )

    service = agents.EventBuilderAgent(
        iam_client=iam_client,
        iter_min_period=3,
    )

    service.start()

    log.info("Bye!!!")


if __name__ == "__main__":
    main()
